Resolves: bz1097904
https://bugzilla.redhat.com/show_bug.cgi?id=1097904

This is back ported from upstream. There's a small conflict in
remap_vmalloc_range.

commit e69e9d4aee712a22665f008ae0550bb3d7c7f7c1
Author: HATAYAMA Daisuke <d.hatay...@jp.fujitsu.com>
Date:   Wed Jul 3 15:02:18 2013 -0700

    vmalloc: introduce remap_vmalloc_range_partial

    We want to allocate ELF note segment buffer on the 2nd kernel in vmalloc
    space and remap it to user-space in order to reduce the risk that memory
    allocation fails on system with huge number of CPUs and so with huge ELF
    note segment that exceeds 11-order block size.

    Although there's already remap_vmalloc_range for the purpose of
    remapping vmalloc memory to user-space, we need to specify user-space
    range via vma.
     Mmap on /proc/vmcore needs to remap range across multiple objects, so
    the interface that requires vma to cover full range is problematic.

    This patch introduces remap_vmalloc_range_partial that receives user-space
    range as a pair of base address and size and can be used for mmap on
    /proc/vmcore case.

    remap_vmalloc_range is rewritten using remap_vmalloc_range_partial.

    [a...@linux-foundation.org: use PAGE_ALIGNED()]
    Signed-off-by: HATAYAMA Daisuke <d.hatay...@jp.fujitsu.com>
    Cc: KOSAKI Motohiro <kosaki.motoh...@jp.fujitsu.com>
    Cc: Vivek Goyal <vgo...@redhat.com>
    Cc: Atsushi Kumagai <kumagai-atsu...@mxc.nes.nec.co.jp>
    Cc: Lisa Mitchell <lisa.mitch...@hp.com>
    Cc: Zhang Yanfei <zhangyan...@cn.fujitsu.com>
    Signed-off-by: Andrew Morton <a...@linux-foundation.org>
    Signed-off-by: Linus Torvalds <torva...@linux-foundation.org>

Signed-off-by: Baoquan He <b...@redhat.com>
---
 include/linux/vmalloc.h |  4 +++
 mm/vmalloc.c            | 68 +++++++++++++++++++++++++++++++++----------------
 2 files changed, 50 insertions(+), 22 deletions(-)

diff --git a/include/linux/vmalloc.h b/include/linux/vmalloc.h
index 56e82f6..1ecf5c7 100644
--- a/include/linux/vmalloc.h
+++ b/include/linux/vmalloc.h
@@ -68,6 +68,10 @@ extern void *vmap(struct page **pages, unsigned int count,
                        unsigned long flags, pgprot_t prot);
 extern void vunmap(const void *addr);
 
+extern int remap_vmalloc_range_partial(struct vm_area_struct *vma,
+                                      unsigned long uaddr, void *kaddr,
+                                      unsigned long size);
+
 extern int remap_vmalloc_range(struct vm_area_struct *vma, void *addr,
                                                        unsigned long pgoff);
 void vmalloc_sync_all(void);
diff --git a/mm/vmalloc.c b/mm/vmalloc.c
index 2ae8618..ff148e0 100644
--- a/mm/vmalloc.c
+++ b/mm/vmalloc.c
@@ -1477,10 +1477,9 @@ static void __vunmap(const void *addr, int 
deallocate_pages)
        if (!addr)
                return;
 
-       if ((PAGE_SIZE-1) & (unsigned long)addr) {
-               WARN(1, KERN_ERR "Trying to vfree() bad address (%p)\n", addr);
+       if (WARN(!PAGE_ALIGNED(addr), "Trying to vfree() bad address (%p)\n",
+                       addr));
                return;
-       }
 
        area = remove_vm_area(addr);
        if (unlikely(!area)) {
@@ -2117,42 +2116,43 @@ finished:
 }
 
 /**
- *     remap_vmalloc_range  -  map vmalloc pages to userspace
- *     @vma:           vma to cover (map full range of vma)
- *     @addr:          vmalloc memory
- *     @pgoff:         number of pages into addr before first page to map
+ *     remap_vmalloc_range_partial  -  map vmalloc pages to userspace
+ *     @vma:           vma to cover
+ *     @uaddr:         target user address to start at
+ *     @kaddr:         virtual address of vmalloc kernel memory
+ *     @size:          size of map area
  *
  *     Returns:        0 for success, -Exxx on failure
  *
  *     This function checks that addr is a valid vmalloc'ed area, and
  *     that it is big enough to cover the vma. Will return failure if
  *     that criteria isn't met.
  *
  *     Similar to remap_pfn_range() (see mm/memory.c)
  */
-int remap_vmalloc_range(struct vm_area_struct *vma, void *addr,
-                                               unsigned long pgoff)
+int remap_vmalloc_range_partial(struct vm_area_struct *vma, unsigned long 
uaddr,
+                               void *kaddr, unsigned long size)
 {
        struct vm_struct *area;
-       unsigned long uaddr = vma->vm_start;
-       unsigned long usize = vma->vm_end - vma->vm_start;
 
-       if ((PAGE_SIZE-1) & (unsigned long)addr)
+       size = PAGE_ALIGN(size);
+
+       if (!PAGE_ALIGNED(uaddr) || !PAGE_ALIGNED(kaddr))
                return -EINVAL;
 
-       area = find_vm_area(addr);
+       area = find_vm_area(kaddr);
        if (!area)
                return -EINVAL;
 
        if (!(area->flags & VM_USERMAP))
                return -EINVAL;
 
-       if (usize + (pgoff << PAGE_SHIFT) > area->size - PAGE_SIZE)
+       if (kaddr + size > area->addr + area->size)
                return -EINVAL;
 
-       addr += pgoff << PAGE_SHIFT;
        do {
-               struct page *page = vmalloc_to_page(addr);
+               struct page *page = vmalloc_to_page(kaddr);
                int ret;
 
                ret = vm_insert_page(vma, uaddr, page);
@@ -2160,15 +2160,39 @@ int remap_vmalloc_range(struct vm_area_struct *vma, 
void *addr,
                        return ret;
 
                uaddr += PAGE_SIZE;
-               addr += PAGE_SIZE;
-               usize -= PAGE_SIZE;
-       } while (usize > 0);
+               kaddr += PAGE_SIZE;
+               size -= PAGE_SIZE;
+       } while (size > 0);
 
        /* Prevent "things" like memory migration? VM_flags need a cleanup... */
        vma->vm_flags |= VM_RESERVED;
 
        return 0;
 }
+EXPORT_SYMBOL(remap_vmalloc_range_partial);
+
+/**
+ *     remap_vmalloc_range  -  map vmalloc pages to userspace
+ *     @vma:           vma to cover (map full range of vma)
+ *     @addr:          vmalloc memory
+ *     @pgoff:         number of pages into addr before first page to map
+ *
+ *     Returns:        0 for success, -Exxx on failure
+ *
+ *     This function checks that addr is a valid vmalloc'ed area, and
+ *     that it is big enough to cover the vma. Will return failure if
+ *     that criteria isn't met.
+ *
+ *     Similar to remap_pfn_range() (see mm/memory.c)
+ */
+int remap_vmalloc_range(struct vm_area_struct *vma, void *addr,
+                                               unsigned long pgoff)
+{
+       return remap_vmalloc_range_partial(vma, vma->vm_start,
+                                          addr + (pgoff << PAGE_SHIFT),
+                                          vma->vm_end - vma->vm_start);
+}
+
 EXPORT_SYMBOL(remap_vmalloc_range);
 
 /*
@@ -2117,42 +2116,43 @@ finished:
  *     Returns:        0 for success, -Exxx on failure
  *
- *     This function checks that addr is a valid vmalloc'ed area, and
- *     that it is big enough to cover the vma. Will return failure if
- *     that criteria isn't met.
+ *     This function checks that @kaddr is a valid vmalloc'ed area,
+ *     and that it is big enough to cover the range starting at
+ *     @uaddr in @vma. Will return failure if that criteria isn't
+ *     met.
  *
  *     Similar to remap_pfn_range() (see mm/memory.c)
-- 
2.1.0

--
To unsubscribe from this list: send the line "unsubscribe linux-kernel" in
the body of a message to majord...@vger.kernel.org
More majordomo info at  http://vger.kernel.org/majordomo-info.html
Please read the FAQ at  http://www.tux.org/lkml/

Reply via email to